body {
	font-family: arial, verdana, sans-serif;
	font-size: 12pt;
	background-color: #FFFFFF;
	color: #000000;
	margin: 0;
	padding: 0.5em;
}

input {
	font-size: 100%;
	vertical-align: middle;
}

input[type="text"] {
	width: 8em;
	max-width: 75%;
	padding: 0.4em;
}

