// generated by tools/gen_alert_header.py

#ifndef LIBTORRENT_ALERTS_H
#define LIBTORRENT_ALERTS_H

// alert types
enum alert_types_t {
	ALERT_TORRENT_ADDED =  3,
	ALERT_TORRENT_REMOVED =  4,
	ALERT_READ_PIECE =  5,
	ALERT_FILE_COMPLETED =  6,
	ALERT_FILE_RENAMED =  7,
	ALERT_FILE_RENAME_FAILED =  8,
	ALERT_PERFORMANCE =  9,
	ALERT_STATE_CHANGED =  10,
	ALERT_TRACKER_ERROR =  11,
	ALERT_TRACKER_WARNING =  12,
	ALERT_SCRAPE_REPLY =  13,
	ALERT_SCRAPE_FAILED =  14,
	ALERT_TRACKER_REPLY =  15,
	ALERT_DHT_REPLY =  16,
	ALERT_TRACKER_ANNOUNCE =  17,
	ALERT_HASH_FAILED =  18,
	ALERT_PEER_BAN =  19,
	ALERT_PEER_UNSNUBBED =  20,
	ALERT_PEER_SNUBBED =  21,
	ALERT_PEER_ERROR =  22,
	ALERT_PEER_CONNECT =  23,
	ALERT_PEER_DISCONNECTED =  24,
	ALERT_INVALID_REQUEST =  25,
	ALERT_TORRENT_FINISHED =  26,
	ALERT_PIECE_FINISHED =  27,
	ALERT_REQUEST_DROPPED =  28,
	ALERT_BLOCK_TIMEOUT =  29,
	ALERT_BLOCK_FINISHED =  30,
	ALERT_BLOCK_DOWNLOADING =  31,
	ALERT_UNWANTED_BLOCK =  32,
	ALERT_STORAGE_MOVED =  33,
	ALERT_STORAGE_MOVED_FAILED =  34,
	ALERT_TORRENT_DELETED =  35,
	ALERT_TORRENT_DELETE_FAILED =  36,
	ALERT_SAVE_RESUME_DATA =  37,
	ALERT_SAVE_RESUME_DATA_FAILED =  38,
	ALERT_TORRENT_PAUSED =  39,
	ALERT_TORRENT_RESUMED =  40,
	ALERT_TORRENT_CHECKED =  41,
	ALERT_URL_SEED =  42,
	ALERT_FILE_ERROR =  43,
	ALERT_METADATA_FAILED =  44,
	ALERT_METADATA_RECEIVED =  45,
	ALERT_UDP_ERROR =  46,
	ALERT_EXTERNAL_IP =  47,
	ALERT_LISTEN_FAILED =  48,
	ALERT_LISTEN_SUCCEEDED =  49,
	ALERT_PORTMAP_ERROR =  50,
	ALERT_PORTMAP =  51,
	ALERT_PORTMAP_LOG =  52,
	ALERT_FASTRESUME_REJECTED =  53,
	ALERT_PEER_BLOCKED =  54,
	ALERT_DHT_ANNOUNCE =  55,
	ALERT_DHT_GET_PEERS =  56,
	ALERT_STATS =  57,
	ALERT_CACHE_FLUSHED =  58,
	ALERT_ANONYMOUS_MODE =  59,
	ALERT_LSD_PEER =  60,
	ALERT_TRACKERID =  61,
	ALERT_DHT_BOOTSTRAP =  62,
	ALERT_TORRENT_ERROR =  64,
	ALERT_TORRENT_NEED_CERT =  65,
	ALERT_INCOMING_CONNECTION =  66,
	ALERT_ADD_TORRENT =  67,
	ALERT_STATE_UPDATE =  68,
	ALERT_MMAP_CACHE =  69,
	ALERT_SESSION_STATS =  70,
	ALERT_DHT_ERROR =  73,
	ALERT_DHT_IMMUTABLE_ITEM =  74,
	ALERT_DHT_MUTABLE_ITEM =  75,
	ALERT_DHT_PUT =  76,
	ALERT_I2P =  77,
	ALERT_DHT_OUTGOING_GET_PEERS =  78,
	ALERT_LOG =  79,
	ALERT_TORRENT_LOG =  80,
	ALERT_PEER_LOG =  81,
	ALERT_LSD_ERROR =  82,
	ALERT_DHT_STATS =  83,
	ALERT_INCOMING_REQUEST =  84,
	ALERT_DHT_LOG =  85,
	ALERT_DHT_PKT =  86,
	ALERT_DHT_GET_PEERS_REPLY =  87,
	ALERT_DHT_DIRECT_RESPONSE =  88,
	ALERT_PICKER_LOG =  89,
	ALERT_SESSION_ERROR =  90,
	ALERT_DHT_LIVE_NODES =  91,
	ALERT_SESSION_STATS_HEADER =  92,
	ALERT_DHT_SAMPLE_INFOHASHES =  93,
	ALERT_BLOCK_UPLOADED =  94,
	ALERT_ALERTS_DROPPED =  95,
	ALERT_SOCKS5 =  96,
	ALERT_FILE_PRIO =  97,
	ALERT_OVERSIZED_FILE =  98,
	ALERT_TORRENT_CONFLICT =  99,
	ALERT_PEER_INFO =  100,
	ALERT_FILE_PROGRESS =  101,
	ALERT_PIECE_INFO =  102,
	ALERT_PIECE_AVAILABILITY =  103,
	ALERT_TRACKER_LIST =  104,
	ALERT_FILE_PRIORITIES =  105,
	ALERT_FILE_STATUS =  106,
};

#endif // LIBTORRENT_ALERTS_H
