/*

Copyright (c) 2014-2015, 2017, 2020-2022, Arvid Norberg
Copyright (c) 2020, Paul-Louis Ageneau
All rights reserved.

You may use, distribute and modify this code under the terms of the BSD license,
see LICENSE file.
*/

#include "test.hpp"
#include "disk_io_test.hpp"
#include "setup_transfer.hpp"
#include "web_seed_suite.hpp"

using namespace lt;

const int proxy = lt::settings_pack::none;

#if TORRENT_USE_SSL
TORRENT_TEST_DISK_IO(url_seed_ssl_keepalive) { run_http_suite(proxy, "https", disk_io); }

TORRENT_TEST_DISK_IO(url_seed_ssl)
{
	run_http_suite(proxy, "https", disk_io, web_seed::no_keepalive);
}
#endif

TORRENT_TEST_DISK_IO(url_seed_keepalive) { run_http_suite(proxy, "http", disk_io); }

TORRENT_TEST_DISK_IO(url_seed) { run_http_suite(proxy, "http", disk_io, web_seed::no_keepalive); }

TORRENT_TEST_DISK_IO(url_seed_host_header)
{
	run_http_suite(proxy, "http", disk_io, web_seed::no_keepalive | web_seed::expect_host_header);
}

TORRENT_TEST_DISK_IO(url_seed_keepalive_rename)
{
	run_http_suite(proxy, "http", disk_io, web_seed::test_ban);
}
