//
// gnutls.hpp
// ~~~~~~~
//
// Copyright (c) 2020 Paul-Louis Ageneau (paul-louis at ageneau dot org)
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
//

#ifndef BOOST_ASIO_GNUTLS_HPP
#define BOOST_ASIO_GNUTLS_HPP

#include <boost/asio/gnutls/context.hpp>
#include <boost/asio/gnutls/context_base.hpp>
#include <boost/asio/gnutls/error.hpp>
#include <boost/asio/gnutls/host_name_verification.hpp>
#include <boost/asio/gnutls/rfc2818_verification.hpp>
#include <boost/asio/gnutls/stream.hpp>
#include <boost/asio/gnutls/stream_base.hpp>
#include <boost/asio/gnutls/verify_context.hpp>

#endif // BOOST_ASIO_GNUTLS_HPP
